//////////////////////////////////////////////////////////////////////////// 
// IPD - Image Processing Design Toolbox
//
// Copyright (c) by Dr. Eng. (J) Harald Galda, 2009 - 2010
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
////////////////////////////////////////////////////////////////////////////


#include "IPD.h"
#include "ScilabToIplImage.h"
#include "IplImageToScilab.h"
#include "TransferImageData.h"
#include "FilterWithMask.h"


template<class TSource, class TDestination>
void FilterImage(TUChar FilterType,
                 const TSource* const Source,
                 TDestination** Destination,
                 TUInt NumberOfColumns, 
                 TUInt NumberOfRows);


template<class TSource, class TDestination>
void SingleFilter(const TSource* const Source,
                  TDestination** Destination,
                  const TSource* const Mask,
                  TUInt NumberOfImageColumns, 
                  TUInt NumberOfImageRows,
                  TUInt NumberOfMaskRows,
                  TUInt NumberOfMaskColumns);


extern "C"
{
 int int_EdgeFilter(char* fname)
 {
  // Input and output parameters are checked.

  CheckRhs(2, 2);

  CheckLhs(1, 1);

  // The first argument is a single channel image of type uint8, uint16, uint32 or double.

  int* ImageAddress = NULL;

  SciErr ScilabError = getVarAddressFromPosition(pvApiCtx, 1, &ImageAddress);

  if(ScilabError.iErr)
	 {
		 printError(&ScilabError, 0);

		 return 0;
	 } // if

  int ImageType = 0;

  int Precision = 0;

  ScilabError = getVarType(pvApiCtx, ImageAddress, &ImageType);

  if(ScilabError.iErr)
	 {
		 printError(&ScilabError, 0);

		 return 0;
	 } // if

  switch(ImageType)
  {
   case sci_matrix:
   {
    // Okay.
   } // case

   break;

   case sci_ints:
   {
    // Only unsigned integer images are processed by IPD toolbox.

    ScilabError = getMatrixOfIntegerPrecision(pvApiCtx, ImageAddress, &Precision);

	   if(ScilabError.iErr)
	   {
		   printError(&ScilabError, 0);

		   return 0;
	   } // if

    switch(Precision)
    {
     case SCI_UINT8: 
     {
      // Okay.
     } // case

     break;

     case SCI_UINT16:
     {
      // Okay.
     } // case

     break;

     case SCI_UINT32:
     {
      // Okay.
     } // case

     break;

     default:
     {
      string Message1("The first argument must be a matrix of type uint8, uint16, ");

      string Message2("uint32 or double.");

      string Message(Message1 + Message2);

      Scierror(ERROR_ARGUMENT, const_cast<char*>(Message.data()));

      return 0;
     } // default

     break;
    } // switch
   } // case

   break;

   default:
   {
    string Message1("The first argument must be a matrix of type uint8, uint16, ");

    string Message2("uint32 or double.");

    string Message(Message1 + Message2);

    Scierror(ERROR_ARGUMENT, const_cast<char*>(Message.data()));

    return 0;
   } // default

   break;
  } // switch

  // The second argument is a scalar of type uint8 which specifies the filter type.

  int* FilterAddress = NULL;

  ScilabError = getVarAddressFromPosition(pvApiCtx, 2, &FilterAddress);

  if(ScilabError.iErr)
	 {
		 printError(&ScilabError, 0);

		 return 0;
	 } // if

  TUChar* FilterType = NULL;

  int NumberOfRows = 0;

  int NumberOfColumns = 0;

  ScilabError = getMatrixOfUnsignedInteger8(pvApiCtx, 
                                            FilterAddress, 
                                            &NumberOfRows,
                                            &NumberOfColumns, 
                                            &FilterType);

  if(ScilabError.iErr)
	 {
		 printError(&ScilabError, 0);

		 return 0;
	 } // if

  if((NumberOfColumns != 1) || (NumberOfRows != 1))
  {
   Scierror(ERROR_SCALAR, "The second argument must be a scalar of type uint8.");

   return 0;
  } // if

  // The image data are retrieved and mapped to a signed data type as follows:
  // uint8 -> int16,  uint16 -> int32, uint32 -> int64, double -> double
  // TUChar -> short, TUShort -> int,  TUInt -> long,  double -> double
  // The signed image data are used to create an image that is filtered.

  switch(ImageType)
  {
   case sci_matrix:
   {
    double* ImageData = NULL;

    ScilabError = getMatrixOfDouble(pvApiCtx, 
                                    ImageAddress, 
                                    &NumberOfRows, 
                                    &NumberOfColumns, 
                                    &ImageData);

    if(ScilabError.iErr)
	   {
		   printError(&ScilabError, 0);

		   return 0;
	   } // if

    double* FilteredImage = NULL;

    FilterImage(*FilterType, ImageData, &FilteredImage, NumberOfColumns, NumberOfRows);

    if(FilteredImage == NULL)
    {
     return 0;
    } // if

    ScilabError = createMatrixOfDouble(pvApiCtx, 
                                       Rhs + 1, 
                                       NumberOfRows, 
                                       NumberOfColumns, 
                                       FilteredImage);

    delete FilteredImage;

    if(ScilabError.iErr)
	   {
		   printError(&ScilabError, 0);

		   return 0;
	   } // if
   } // case

   break;

   case sci_ints:
   {
    switch(Precision)
    {
     case SCI_UINT8:
     {
      TUChar* ImageData = NULL;

      ScilabError = getMatrixOfUnsignedInteger8(pvApiCtx, 
                                                ImageAddress, 
                                                &NumberOfRows, 
                                                &NumberOfColumns, 
                                                &ImageData);

      if(ScilabError.iErr)
	     {
		     printError(&ScilabError, 0);

		     return 0;
	     } // if

      short* SignedImageData = NULL;

      TransferImageData(NumberOfColumns * NumberOfRows, ImageData, &SignedImageData);

      if(SignedImageData == NULL)
      {
       Scierror(ERROR_MEMORY, "Internal error occured.");

       return 0;
      } // if

      TUChar * FilteredImage = NULL;
      
      FilterImage(*FilterType, SignedImageData, &FilteredImage, NumberOfColumns, NumberOfRows);

      if(FilteredImage == NULL)
      {
       return 0;
      } // if

      ScilabError = createMatrixOfUnsignedInteger8(pvApiCtx, 
                                                   Rhs + 1, 
                                                   NumberOfRows, 
                                                   NumberOfColumns, 
                                                   FilteredImage);

      delete FilteredImage;

      if(ScilabError.iErr)
	     {
		     printError(&ScilabError, 0);

		     return 0;
	     } // if
     } // case

     break;

     case SCI_UINT16:
     {
      TUShort* ImageData = NULL;

      ScilabError = getMatrixOfUnsignedInteger16(pvApiCtx, 
                                                 ImageAddress, 
                                                 &NumberOfRows, 
                                                 &NumberOfColumns, 
                                                 &ImageData);

      if(ScilabError.iErr)
	     {
		     printError(&ScilabError, 0);

		     return 0;
	     } // if

      int* SignedImageData = NULL;

      TransferImageData(NumberOfColumns * NumberOfRows, ImageData, &SignedImageData);

      if(SignedImageData == NULL)
      {
       Scierror(ERROR_MEMORY, "Internal error occured.");

       return 0;
      } // if

      TUShort* FilteredImage = NULL;

      FilterImage(*FilterType, SignedImageData, &FilteredImage, NumberOfColumns, NumberOfRows);

      if(FilteredImage == NULL)
      {
       return 0;
      } // if

      ScilabError = createMatrixOfUnsignedInteger16(pvApiCtx, 
                                                    Rhs + 1, 
                                                    NumberOfRows, 
                                                    NumberOfColumns, 
                                                    FilteredImage);

      delete FilteredImage;

      if(ScilabError.iErr)
	     {
		     printError(&ScilabError, 0);

		     return 0;
	     } // if
     } // case

     break;

     case SCI_UINT32:
     {
      TUInt* ImageData = NULL;

      ScilabError = getMatrixOfUnsignedInteger32(pvApiCtx, 
                                                 ImageAddress, 
                                                 &NumberOfRows, 
                                                 &NumberOfColumns, 
                                                 &ImageData);

      if(ScilabError.iErr)
	     {
		     printError(&ScilabError, 0);

		     return 0;
	     } // if

      long* SignedImageData = NULL;

      TransferImageData(NumberOfColumns * NumberOfRows, ImageData, &SignedImageData);

      if(SignedImageData == NULL)
      {
       Scierror(ERROR_MEMORY, "Internal error occured.");

       return 0;
      } // if

      TUInt* FilteredImage = NULL;

      FilterImage(*FilterType, SignedImageData, &FilteredImage, NumberOfColumns, NumberOfRows);

      if(FilteredImage == NULL)
      {
       return 0;
      } // if

      ScilabError = createMatrixOfUnsignedInteger32(pvApiCtx, 
                                                    Rhs + 1, 
                                                    NumberOfRows, 
                                                    NumberOfColumns, 
                                                    FilteredImage);

      delete FilteredImage;

      if(ScilabError.iErr)
	     {
		     printError(&ScilabError, 0);

		     return 0;
	     } // if
     } // case

     break;
    } // switch
   } // case
  
   break;

   default:
   {
   } // default

   break;
  } // switch

  // The output argument position is set.
  LhsVar(1) = Rhs + 1;

  return 0;
 } // int_EdgeFilter
}


template<class TSource, class TDestination>
void FilterImage(TUChar FilterType,
                 const TSource* const Source,
                 TDestination** Destination,
                 TUInt NumberOfColumns, 
                 TUInt NumberOfRows)
{
 // initializations

 TUInt NumberOfPixels = NumberOfColumns * NumberOfRows;

 *Destination = NULL;

 TDestination* Result = NULL;

 const TUChar NUMBER_OF_COEFFICIENTS = 9;

 TSource Mask[NUMBER_OF_COEFFICIENTS];

 switch(FilterType)
 {
  case EDGE_SOBEL:
  {
   // The result image for horizontal gradient component is initialized.
   TDestination* Horizontal = NULL;

   // The mask for the horizontal gradient component is initialized.

   Mask[0] = -1;

   Mask[1] = -2;

   Mask[2] = -1;

   Mask[3] = 0;

   Mask[4] = 0;

   Mask[5] = 0;

   Mask[6] = 1;

   Mask[7] = 2;

   Mask[8] = 1;

   // The horizontal gradient component is calculated.

   FilterWithMask(Source, NumberOfColumns, NumberOfRows, Mask, 3, 3, &Horizontal);

   if(Horizontal == NULL)
   {
    return;
   } // if

   // The image for the vertical gradient component is initialized.
   TDestination* Vertical = NULL;

   // The mask for the vertical gradient component is initialized.

   Mask[0] = -1;

   Mask[1] = 0;

   Mask[2] = 1;

   Mask[3] = -2;

   Mask[4] = 0;

   Mask[5] = 2;

   Mask[6] = -1;

   Mask[7] = 0;

   Mask[8] = 1;

   // The vertical gradient component is calculated.

   FilterWithMask(Source, NumberOfColumns, NumberOfRows, Mask, 3, 3, &Vertical);

   if(Horizontal == NULL)
   {
    return;
   } // if

   // The absolute value of the gradient is approximated by the sum of the absolute values of
   // horizontal and vertical components for each pixel. The result ist stored
   // in Horizontal in order to save memory.

   for(TUInt n = 0; n < NumberOfPixels; n++)
   {
    Horizontal[n] = Horizontal[n] + Vertical[n];
   } // for

   delete Vertical;

   Result = Horizontal;
  } // case

  break;

  case EDGE_LAPLACE:
  {
   // Mask is initalized.
   
   for(TUChar n = 0; n < NUMBER_OF_COEFFICIENTS; n++)
   {
    if(n != 4)
    {
     Mask[n] = 1;
    }
    else
    {
     Mask[n] = -8;
    } // if
   } // for

   // Mask is applied.
   FilterWithMask(Source, NumberOfColumns, NumberOfRows, Mask, 3, 3, &Result);
  } // case

  break;

  case EDGE_PREWITT:
  {
   // The result image for horizontal gradient component is initialized.
   TDestination* Horizontal = NULL;

   // The mask for the horizontal gradient component is initialized.

   Mask[0] = -1;

   Mask[1] = -1;

   Mask[2] = -1;

   Mask[3] = 0;

   Mask[4] = 0;

   Mask[5] = 0;

   Mask[6] = 1;

   Mask[7] = 1;

   Mask[8] = 1;

   // The horizontal gradient component is calculated.

   FilterWithMask(Source, NumberOfColumns, NumberOfRows, Mask, 3, 3, &Horizontal);

   if(Horizontal == NULL)
   {
    return;
   } // if

   // The image for the vertical gradient component is initialized.
   TDestination* Vertical = NULL;

   // The mask for the vertical gradient component is initialized.

   Mask[0] = -1;

   Mask[1] = 0;

   Mask[2] = 1;

   Mask[3] = -1;

   Mask[4] = 0;

   Mask[5] = 1;

   Mask[6] = -1;

   Mask[7] = 0;

   Mask[8] = 1;

   // The vertical gradient component is calculated.

   FilterWithMask(Source, NumberOfColumns, NumberOfRows, Mask, 3, 3, &Vertical);

   if(Horizontal == NULL)
   {
    return;
   } // if

   // The absolute value of the gradient is approximated by the sum of the absolute values of
   // horizontal and vertical components for each pixel. The result ist stored
   // in Horizontal in order to save memory.

   for(TUInt n = 0; n < NumberOfPixels; n++)
   {
    Horizontal[n] = Horizontal[n] + Vertical[n];
   } // for

   delete Vertical;

   Result = Horizontal;
  } // case

  case EDGE_SCHARR:
  {
      // The result image for horizontal gradient component is initialized.
   TDestination* Horizontal = NULL;

   // The mask for the horizontal gradient component is initialized.

   Mask[0] = -3;

   Mask[1] = -10;

   Mask[2] = -3;

   Mask[3] = 0;

   Mask[4] = 0;

   Mask[5] = 0;

   Mask[6] = 3;

   Mask[7] = 10;

   Mask[8] = 3;

   // The horizontal gradient component is calculated.

   FilterWithMask(Source, NumberOfColumns, NumberOfRows, Mask, 3, 3, &Horizontal);

   if(Horizontal == NULL)
   {
    return;
   } // if

   // The image for the vertical gradient component is initialized.
   TDestination* Vertical = NULL;

   // The mask for the vertical gradient component is initialized.

   Mask[0] = -3;

   Mask[1] = 0;

   Mask[2] = 3;

   Mask[3] = -10;

   Mask[4] = 0;

   Mask[5] = 10;

   Mask[6] = -3;

   Mask[7] = 0;

   Mask[8] = 3;

   // The vertical gradient component is calculated.

   FilterWithMask(Source, NumberOfColumns, NumberOfRows, Mask, 3, 3, &Vertical);

   if(Horizontal == NULL)
   {
    return;
   } // if

   // The absolute value of the gradient is approximated by the sum of the absolute values of
   // horizontal and vertical components for each pixel. The result ist stored
   // in Horizontal in order to save memory.

   for(TUInt n = 0; n < NumberOfPixels; n++)
   {
    Horizontal[n] = Horizontal[n] + Vertical[n];
   } // for

   delete Vertical;

   Result = Horizontal;
  } // case

  break;

  default:
  {
   Scierror(ERROR_ARGUMENT, "Unknown filter type passed to second argument.");

   return;
  } // default

  break;
 } // switch

 if(Result == NULL)
 {
  Scierror(ERROR_MEMORY, "Internal error occured.");

  return;
 } // if

 *Destination = Result;
} // FilterImage
