//////////////////////////////////////////////////////////////////////////// 
// IPD - Image Processing Design Toolbox
//
// Copyright (c) by Dr. Eng. (J) Harald Galda, 2009 - 2010
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
////////////////////////////////////////////////////////////////////////////

#ifndef FILTER_H
#define FILTER_H

#include "IPD.h"


template<class GrayValueType>
class CFilter
{
 public:

  CFilter(TUInt NumberOfColumns, TUInt NumberOfRows);

  virtual ~CFilter(){};

  virtual void FilterImage(TUInt ImageWidth, 
                           TUInt ImageHeight, 
                           const GrayValueType* const Source, 
                           GrayValueType* Destination);

 protected:

  typedef vector<GrayValueType> TGrayValueList;

  typedef vector<int> TOffsetList;

  virtual GrayValueType ComputeGrayValue(const TGrayValueList& GrayValues) = 0;

  virtual void ComputeOffsets(TUInt NumberOfColumns, TUInt NumberOfRows);

  TUInt SubscriptsToIndex(TUInt ImageHeight, TUInt Column, TUInt Row);

  TUInt m_NumberOfElements;
  
  TOffsetList m_ColumnOffsets;

  TOffsetList m_RowOffsets;

 private:

  CFilter():m_Values(NULL){};

  CFilter(const CFilter&):m_Values(NULL){};
}; // CFilter


template<class GrayValueType>
CFilter<GrayValueType>::CFilter(TUInt NumberOfColumns, TUInt NumberOfRows):
m_NumberOfElements(NumberOfColumns * NumberOfRows),
m_ColumnOffsets(TOffsetList(0)),
m_RowOffsets(TOffsetList(0))
{
} // CFilter::CFilter


template<class GrayValueType>
void CFilter<GrayValueType>::FilterImage(TUInt ImageWidth, 
                                         TUInt ImageHeight, 
                                         const GrayValueType* const Source, 
                                         GrayValueType* Destination)
{
 // If Source and Destination point to the same address, a different array must be alloca-
 // ted. If this fails, the function has no effect.

 bool InPlace = (Source == Destination);

 GrayValueType* Temp = NULL;

 if(InPlace)
 {
  Temp = new GrayValueType[ImageWidth * ImageHeight];

  if(Temp == NULL)
  {
   return;
  } // if
 } // if

 TUInt PixelIndex = 0;

 TGrayValueList GrayValues(m_NumberOfElements);

 for(TUInt Column = 0; Column < ImageWidth; Column++)
 {
  for(TUInt Row = 0; Row < ImageHeight; Row++)
  {
   // For the current pixel a list of gray values at its neighbor pixels is calculated. A
   // gray value is selected from this list and stored in Destination.

   for(TUInt Index = 0; Index < m_NumberOfElements; Index++)
   {
    TUInt NeighborColumn = 0;

    long NewColumn = Column + m_ColumnOffsets[Index];

    if(NewColumn >= 0)
    {
     NeighborColumn = NewColumn;

     if(NeighborColumn >= ImageWidth)
     {
      NeighborColumn = ImageWidth - 1 - (NeighborColumn - ImageWidth);
     }// if
    }
    else
    {
     NeighborColumn = -NewColumn - 1;
    }// if

    TUInt NeighborRow = 0;

    long NewRow = Row + m_RowOffsets[Index];

    if(NewRow >= 0)
    {
     NeighborRow = NewRow;

     if(NeighborRow >= ImageHeight)
     {
      NeighborRow = ImageHeight - 1 - (NeighborRow - ImageHeight);
     }// if
    }
    else
    {
     NeighborRow = -NewRow - 1;
    }// if

    GrayValues[Index] = Source[SubscriptsToIndex(ImageHeight, NeighborColumn, NeighborRow)];
   } // for

   if(InPlace)
   {
    Temp[PixelIndex] = ComputeGrayValue(GrayValues);
   }
   else
   {
    Destination[PixelIndex] = ComputeGrayValue(GrayValues);
   } // if

   PixelIndex++;
  } // for
 } // for

 // If Source and Destination point to the same array, Temp is copied to Destination and 
 // Temp is released.
 if(InPlace)
 {
  memcpy(Destination, Temp, ImageWidth * ImageHeight * sizeof(GrayValueType));

  delete Temp;
 } // if
} // CFilter::FilterImage


template<class GrayValueType>
void CFilter<GrayValueType>::ComputeOffsets(TUInt NumberOfColumns, TUInt NumberOfRows)
{
 // Central pixel coordinates are calculated.

 TUInt CentralColumn = NumberOfColumns / 2;

 TUInt CentralRow = NumberOfRows / 2;

 // For each non-zero pixel of the filter the horizontal and vertical distances to the cen- 
 // tral pixel are calculated and stored in m_ColumnOffsets and m_RowOffsets respectively. 

 TUInt ElementIndex = 0;

 for(TUInt n = 0; n < NumberOfColumns; n++)
 {
  for(TUInt m = 0; m < NumberOfRows; m++)
  {
   m_ColumnOffsets.push_back(n - CentralColumn);

   m_RowOffsets.push_back(m - CentralRow);
  } // for
 } // for
} // CFilter::ComputeOffsets


template<class GrayValueType>
TUInt CFilter<GrayValueType>::SubscriptsToIndex(TUInt ImageHeight, TUInt Column, TUInt Row)
{
 return Column * ImageHeight + Row;
} // CFilter::SubscriptsToIndex


#endif